import enum
import os

# Used to import null _pylumi name on ReadTheDocs
try:
    import _pylumi
    from _pylumi import (
        UNKNOWN_KEY,
        UNKNOWN_BOOL_VALUE,
        UNKNOWN_NUMBER_VALUE,
        UNKNOWN_STRING_VALUE,
        UNKNOWN_ARRAY_VALUE,
        UNKNOWN_ASSET_VALUE,
        UNKNOWN_ARCHIVE_VALUE,
        UNKNOWN_OBJECT_VALUE,
        DIFF_ADD,
        DIFF_ADD_REPLACE,
        DIFF_DELETE,
        DIFF_DELETE_REPLACE,
        DIFF_UPDATE,
        DIFF_UPDATE_REPLACE,
    )

except ImportError:
    if not os.getenv("READTHEDOCS"):
        raise

    _pylumi = None

    UNKNOWN_KEY = None
    UNKNOWN_BOOL_VALUE = None
    UNKNOWN_NUMBER_VALUE = None
    UNKNOWN_STRING_VALUE = None
    UNKNOWN_ARRAY_VALUE = None
    UNKNOWN_ASSET_VALUE = None
    UNKNOWN_ARCHIVE_VALUE = None
    UNKNOWN_OBJECT_VALUE = None

    DIFF_ADD = None
    DIFF_ADD_REPLACE = None
    DIFF_DELETE = None
    DIFF_DELETE_REPLACE = None
    DIFF_UPDATE = None
    DIFF_UPDATE_REPLACE = None


class UnknownValue(enum.Enum):
    """
    Enum for the UNKNOWN_*_VALUE values
    """

    BOOL = UNKNOWN_BOOL_VALUE
    NUMBER = UNKNOWN_NUMBER_VALUE
    STRING = UNKNOWN_STRING_VALUE
    ARRAY = UNKNOWN_ARRAY_VALUE
    ASSET = UNKNOWN_ASSET_VALUE
    ARCHIVE = UNKNOWN_ARCHIVE_VALUE
    OBJECT = UNKNOWN_OBJECT_VALUE


class DiffType(enum.Enum):
    """
    Enum for diff types
    """

    ADD = DIFF_ADD
    ADD_REPLACE = DIFF_ADD_REPLACE
    DELETE = DIFF_DELETE
    DELETE_REPLACE = DIFF_DELETE_REPLACE
    UPDATE = DIFF_UPDATE
    UPDATE_REPLACE = DIFF_UPDATE_REPLACE



from pylumi import exc
from pylumi.context import Context
from pylumi.ext import (
    UNKNOWN_KEY,
    UNKNOWN_BOOL_VALUE,
    UNKNOWN_NUMBER_VALUE,
    UNKNOWN_STRING_VALUE,
    UNKNOWN_ARRAY_VALUE,
    UNKNOWN_ASSET_VALUE,
    UNKNOWN_ARCHIVE_VALUE,
    UNKNOWN_OBJECT_VALUE,
    DIFF_ADD,
    DIFF_ADD_REPLACE,
    DIFF_DELETE,
    DIFF_DELETE_REPLACE,
    DIFF_UPDATE,
    DIFF_UPDATE_REPLACE,
    UnknownValue,
    DiffType,
)
from pylumi.provider import Provider
from pylumi.urn import URN

__version__ = "1.1.9"

__pulumi_version__ = "2.12.0"



import dataclasses as dc
import re

from pylumi import exc


URN_TEMPLATE = "urn:pulumi:{stack}::{project}::{type}::{name}"


URN_PATTERNS = (
    r"^urn:pulumi:(?P<stack>[a-zA-Z0-9_\-/]+?)::(?P<project>[a-zA-Z0-9_\-/]+?)"
    r"::(?P<type>[a-zA-Z0-9_\-/:]+?)::(?P<name>[a-zA-Z0-9_\-/]+?)$",
    # A short form of a URN omitting project and stack info.
    r"^urn:pulumi:(?P<type>[a-zA-Z0-9_\-/:]+?)::(?P<name>[a-zA-Z0-9_\-/]+?)$",
)


BLANK = "_"


@dc.dataclass(frozen=True, repr=False)
class URN:
    """
    A URN builder class
    """

    type: str
    name: str = BLANK
    stack: str = BLANK
    project: str = BLANK

    @classmethod
    def parse(cls, urn: str) -> "URN":
        """
        Construct a URN object from a string. Does not need to be called directly,
        a string can be passed as a positional argument to URN e.g. URN('urn:pulumi...')
        """
        for pattern in URN_PATTERNS:
            srch = re.search(pattern, urn)
            if srch is not None:
                groups = srch.groupdict()
                return cls(**groups)

        raise exc.InvalidURN(urn)

    def __post_init__(self) -> None:
        if self.type.startswith("urn:pulumi"):
            obj = type(self).parse(self.type)
            for key, val in dc.asdict(obj).items():
                if self.__dict__[key] == BLANK or key == "type":
                    self.__dict__[key] = val

    def render(self) -> str:
        """
        Render this URN object to a string
        """
        return URN_TEMPLATE.format_map(dc.asdict(self))

    def replace(self, **kwargs) -> "URN":
        """
        Return a new URN with the given components replaced
        """
        return dc.replace(self, **kwargs)

    def __repr__(self) -> str:
        """
        Render an object representation for this URN.
        """
        return f"URN({self.render()})"

    def __str__(self) -> str:
        """
        Render a string representation of this URN.
        """
        return self.render()



import os
import uuid
from functools import partial, wraps
from typing import Any, Sequence, Optional, Dict

from pylumi import provider
from pylumi.ext import _pylumi


class Context:
    """
    A context is a Python representation of a statey plugin context, which
    manages resource plugins. Basically this acts as an orchestration server
    which exposes a gRPC server and proxies communications with the actual
    resource plugin processes (my understandning).

    **Parameters:**

    * **name** - (optional) Assign a name to the context; this name is unique,
    so if two contexts are created with the same name then they will point to
    the same Provider object in the go runtime.
    * **cwd** - (optional) Pass a current working directory to use for the context.

    """

    def __init__(self, name: Optional[str] = None, cwd: Optional[str] = None) -> None:
        if cwd is None:
            cwd = os.getcwd()
        if name is None:
            name = uuid.uuid4().hex

        self.name = name
        self.cwd = cwd

    def provider(
        self, name: str, config: Optional[Dict[str, Any]] = None
    ) -> provider.Provider:
        """
        Get a Provider object with the given name. This just creates the provider object,
        no interaction is done with the Pulumi engine until configure() if called (or
        the provider is used as a context manager).

        **Parameters:**

        * **name** - The name of the provider, e.g. 'aws'.
        * **config** - (optional) configuration parameters for the provider.

        **Returns:**

        A new Provider instance.

        **Pulumi Docs:**

        Provider loads a new copy of the provider for a given package.  If a provider for
        this package could not be found, or an error occurs while creating it, a non-nil
        error is returned.
        Reference: `Provider <github.com/pulumi/pulumi/sdk/v2/go/common/resource/plugin/host.go>`_
        """
        if config is None:
            config = {}
        return provider.Provider(self, name, config)

    def setup(self) -> None:
        """
        Set up this Pulumi context. This creates an interface in the Go runtime
        that can create and communicate with resource provider proce

        **Returns:**

        None
        """
        return _pylumi.context_setup(self.name, self.cwd)

    def teardown(self) -> None:
        """
        Tear down this provider, removing associated OS resources such as plugin
        processes.

        **Returns:**

        None
        """
        return _pylumi.context_teardown(self.name)

    def list_plugins(self) -> Sequence[str]:
        """
        List the currently loaded plugins in this context.

        **Returns:**

        A list of plugin names that are currently loaded in the context.

        **Pulumi docs**:

        ListPlugins lists all plugins that have been loaded, with version information.
        Reference: `ListPlugins <github.com/pulumi/pulumi/sdk/v2/go/common/resource/plugin/host.go>`_
        """
        return _pylumi.context_list_plugins(self.name)

    def __enter__(self) -> Any:
        self.setup()
        return self

    def __exit__(self, exc_type, exc_value, tb) -> None:
        self.teardown()



class PylumiError(Exception):
    """
    Base class for Pylumi errors
    """


class InvalidURN(PylumiError):
    """
    Error from parsing a URN value.
    """

    def __init__(self, value: str) -> None:
        self.value = value
        super().__init__(f"Invalid URN value: {repr(value)}.")
